/*

@author Sergio Alestra

*/

import ij.*;
import ij.gui.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import ij.text.*;

//plugin che si occupa di migliorare il contrasto

/*
IF un pixel  scuro  THEN rendilo pi scuro
IF un pixel  grigio THEN lascialo grigio
IF un pixel  chiaro THEN rendilo pi chiaro
*/


public class Fuzzy_Contrast_Enhancement extends WindowAdapter implements PlugInFilter, ChangeListener, ActionListener{
  
  private int w;
  private int h;
  private ImagePlus im_fce;
  private ImageProcessor ip_orig,ip_fce;
  private byte[] im;
  private ImageWindow w_fce;
  
  //variabili per le regole
  private String[] rules;
  private TextWindow windowRules;
  
  //variabili per le funzioni di membership
  private float aDark = 70f;
  private float bDark = 57f;
  private float aBright = 180f;
  private float bBright = 57f;
  private float aGray = 127f;
  private float bGray = 57f;
  private float cGray = 57f;
  
  //Funzioni Singleton
  private int DARKER = 0;
  private int GRAY = 127;
  private int BRIGHTER = 255;
  
  //slider per input mbf
  private JSlider s_aDark;
  private JSlider s_bDark;
  private JSlider s_aGray;
  private JSlider s_bGray;
  private JSlider s_cGray;
  private JSlider s_aBright;
  private JSlider s_bBright;
  
  //slider per output singleton
  private JSlider s_DARKER;
  private JSlider s_GRAY;
  private JSlider s_BRIGHTER;
  
  //etichette per input mbf
  private JLabel l_aDark;
  private JLabel l_bDark;
  private JLabel l_aGray;
  private JLabel l_bGray;
  private JLabel l_cGray;
  private JLabel l_aBright;
  private JLabel l_bBright;
  
  //etichette per output singleton
  private JLabel l_DARKER;
  private JLabel l_GRAY;
  private JLabel l_BRIGHTER;
  
  //bottonI applyMbf  resetMbf
  private JButton applyMbf,resetMbf;
  
  //bottone applySingleton e resetSingleton
  private JButton applySingleton,resetSingleton;
  
  //finestra per visualizzare le mbf
  private JFrame windowMbf;
  private PlotPanel graphicMbf;
  
  //finestra per visualizzare i singleton
  private JFrame windowSingleton;
  private PlotPanel graphicSingleton;
  
  
  //metodo di inizializzazione richiesto da PlugInFilter
  public int setup(String arg, ImagePlus imp) {
    
    if (IJ.versionLessThan("1.37j"))
			return DONE;
		else
			return DOES_8G+NO_CHANGES;
	}
  
  //metodo di esecuzione richiesto da PlugInFilter
  public void run(ImageProcessor ip) {
    
    w = ip.getWidth();
    h = ip.getHeight();
    ip_orig=ip;
    
    
    //creazione di una copia dell'immagine su cui lavorare
    im_fce = NewImage.createByteImage("Fuzzy Contrast Enhancement",w,h,1,NewImage.FILL_BLACK);
    ip_fce = (im_fce.getProcessor()).convertToByte(true);
    ip_fce.copyBits(ip,0,0,Blitter.COPY);
      
      
    //recuperiamo i valori dei pixels dell'immagine
    im = (byte[]) ip_fce.getPixels();
      
    //processiamo l'immagine
    this.process();
    
    //creiamo e visualizziamo le regole
    this.makeRules();
    this.showRules();
    
    //visualizziamo le mbf
    this.showMbf();
    this.showSingleton();
    
    //visualizziamo l'immagine
    w_fce = new ImageWindow(im_fce);
    w_fce.addWindowListener(this);
    im_fce.updateAndDraw();
	}
  
  //-----------metodo makeRules-------------
  private void makeRules(){
    
    rules = new String[3];
    rules[0] = "IF a pixel is dark, THEN make it darker";
    rules[1] = "IF a pixel is gray, THEN make it gray";
    rules[2] = "IF a pixel is bright, THEN make it brighter";
  }
  
  //metodo per mostrare le regole
  private void showRules(){
    
    windowRules = new TextWindow("Rules - Fuzzy Contrast Enhancement",
                                  "", 500, 200);
    
    for(int i=0; i<rules.length; i++)
      windowRules.append(rules[i]);
    
    windowRules.setVisible(true);
  }
  
  //---------metodo process--------------
  private void process(){
    
    int pos, offset, int_im_pos;
    float dark_fuzzy_px, gray_fuzzy_px, bright_fuzzy_px;
    
    //scandisco tutti i pixels
    for(int i = 0; i < h; i++){
      
      offset = i*w;
      for(int j = 0; j < w; j++){
        
        pos = offset+j;
        
        //cast ad int del valore di input del pixel
        int_im_pos = 0xff & im[pos];
        
        //fuzzyfichiamo l'input
        dark_fuzzy_px = this.mbfDark(int_im_pos);
        gray_fuzzy_px = this.mbfGray(int_im_pos);
        bright_fuzzy_px = this.mbfBright(int_im_pos);
        
        //calcoliamo il valore di output del pixel
        int_im_pos = (int) ((dark_fuzzy_px*DARKER + 
                          gray_fuzzy_px*GRAY + 
                          bright_fuzzy_px*BRIGHTER)/
                          (dark_fuzzy_px + gray_fuzzy_px + bright_fuzzy_px));
        
        //cast a byte del valore di output del pixel
        
        im[pos] = (byte) int_im_pos;
      }
    }
  }
  
   
  
  //funzione di Membership Scuro di tipo Sigma
  private float mbfDark(int px){
    
    if( (px >= (aDark)) && (px <= (aDark+bDark)))
      return 1-(px-aDark)/bDark;
    else if( px < aDark )
      return 1;
    else
      return 0;
  }
  
  //funzione di Membership Chiaro di tipo Sigma
  private float mbfBright(int px){
    
    if( (px >= (aBright-bBright)) && (px < aBright))
      return 1-(aBright-px)/bBright;
    else if( px >= aBright )
      return 1;
    else
      return 0;
  }
  
  //funzione di Membership Grigio di tipo Triangolo
  private float mbfGray(int px){
    
    if( (px >= (aGray-bGray)) && (px < aGray) )
      return 1-(aGray-px)/bGray;
    else if( (px >= aGray) && (px <= (aGray+cGray)) )
      return 1-(px-aGray)/cGray;
    else
      return 0;
  }
  
  //metodo per plottare le mbf
  private ImagePlus plotMbf(){
    
    
    float[] x = new float[256];
    float[] yD = new float[256];
    float[] yG = new float[256];
    float[] yB = new float[256];
    
    for(int i =0; i<256; i++){
      x[i]=i;
      yD[i]=mbfDark(i);
      yG[i]=mbfGray(i);
      yB[i]=mbfBright(i);
    }
    
    Plot p = new Plot("Membership function - Fuzzy Contrast Enhancement",
                                    "","",x,yD);
    p.setLimits(0.0,255.0,0.0,1.2);
    p.setLineWidth(2);
    
    p.addPoints(x,yG,Plot.LINE);
    p.addPoints(x,yB,Plot.LINE);
    
    //p.addLabel(0.1,0.1,"mbf Dark");
    //p.addLabel(0.45,0.1,"mbf Gray");
    //p.addLabel(0.8,0.1,"mbf Bright");
    
    return p.getImagePlus();
    
  }
  
  //metodo per plottare i singleton
  private ImagePlus plotSingleton(){
    
    
    float[] xD = new float[100];
    float[] xG = new float[100];
    float[] xB = new float[100];
    float[] y = new float[100];
    for(int i=0; i<100; i++){
      xD[i]=DARKER;
      xG[i]=GRAY;
      xB[i]=BRIGHTER;
      y[i]=i*0.01f;
    }
    
    Plot p = new Plot("Singleton - Fuzzy Contrast Enhancement",
                                    "","",xD,y);
    p.setLimits(0.0,255.0,0.0,1.2);
    p.setLineWidth(2);
    p.addPoints(xG,y,PlotWindow.LINE);
    p.addPoints(xB,y,PlotWindow.LINE);
    
    //p.addLabel(0,0.2,"Darker");
    //p.addLabel(0.4,0.2,"Gray");
    //p.addLabel(0.9,0.2,"Brighter");
    
    return p.getImagePlus();
  }
  
  //--------------metodi per la finestra delle Mbf
  
  private void showMbf(){
    
    //creo i bottoni
    applyMbf = new JButton("Apply");
    applyMbf.addActionListener(this);
    resetMbf = new JButton("Reset");
    resetMbf.addActionListener(this);
    
    //creo i pannelli
    JPanel panelDark = new JPanel(new GridLayout(2,2));
    JPanel panelBright = new JPanel(new GridLayout(2,2));
    JPanel panelGray = new JPanel(new GridLayout(2,3));
    JPanel panelApply_Reset = new JPanel();
    graphicMbf = new PlotPanel(this.plotMbf().getImage());
    
    //aggiungo un bordo ai pannelli
    panelDark.setBorder(BorderFactory.createTitledBorder("Membership function Dark"));
    panelGray.setBorder(BorderFactory.createTitledBorder("Membership function Gray"));
    panelBright.setBorder(BorderFactory.createTitledBorder("Membership function Bright"));
    
    //creo le etichette
    l_aDark = new JLabel();
    l_bDark = new JLabel();
    l_aGray = new JLabel();
    l_bGray = new JLabel();
    l_cGray = new JLabel();
    l_aBright = new JLabel();
    l_bBright = new JLabel();
    
    this.setLabelMbf();
    
    //aggiungo le etichette dei campi
    panelDark.add(l_aDark);
    panelDark.add(l_bDark);
    panelGray.add(l_aGray);
    panelGray.add(l_bGray);
    panelGray.add(l_cGray);
    panelBright.add(l_aBright);
    panelBright.add(l_bBright);
    
    //creo gli slider
    s_aDark = new JSlider(JSlider.HORIZONTAL);
    s_bDark = new JSlider(JSlider.HORIZONTAL);
    s_aGray = new JSlider(JSlider.HORIZONTAL);
    s_bGray = new JSlider(JSlider.HORIZONTAL);
    s_cGray = new JSlider(JSlider.HORIZONTAL);
    s_aBright = new JSlider(JSlider.HORIZONTAL);
    s_bBright = new JSlider(JSlider.HORIZONTAL);
    
    this.setSliderMbf();
    
    s_aDark.addChangeListener(this);
    s_bDark.addChangeListener(this);
    s_aGray.addChangeListener(this);
    s_bGray.addChangeListener(this);
    s_cGray.addChangeListener(this);
    s_aBright.addChangeListener(this);
    s_bBright.addChangeListener(this);
    
    //aggiungo gli slider
    panelDark.add(s_aDark);
    panelDark.add(s_bDark);
    panelGray.add(s_aGray);
    panelGray.add(s_bGray);
    panelGray.add(s_cGray);
    panelBright.add(s_aBright);
    panelBright.add(s_bBright);
    panelApply_Reset.add(applyMbf);
    panelApply_Reset.add(resetMbf);
    
    //creo la finestra per le mbf
    windowMbf = new JFrame("Membership function - Fuzzy Contrast Enhancement");
    windowMbf.setSize(700,550);
    windowMbf.setLocation(300,200);
    windowMbf.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    windowMbf.setLayout(new FlowLayout());
    Container contentPane=windowMbf.getContentPane();
    
    contentPane.add(graphicMbf);
    contentPane.add(panelDark);
    contentPane.add(panelGray);
    contentPane.add(panelBright);
    contentPane.add(panelApply_Reset);
    
    windowMbf.setVisible(true);
  }
  
  private void setLabelMbf(){
    
    l_aDark.setText("a = "+(int)aDark);
    l_bDark.setText("b = "+(int)bDark);
    l_aGray.setText("a = "+(int)aGray);
    l_bGray.setText("b = "+(int)bGray);
    l_cGray.setText("c = "+(int)cGray);
    l_aBright.setText("a = "+(int)aBright);
    l_bBright.setText("b = "+(int)bBright);
  }
  
  private void setSliderMbf(){
    
    s_aDark.setMinimum(0);
    s_aDark.setMaximum(255-(int)bDark);
    s_aDark.setValue((int)aDark);
    s_bDark.setMinimum(0);
    s_bDark.setMaximum(255-(int)aDark);
    s_bDark.setValue((int)bDark);
    
    s_aBright.setMinimum(0+(int)bBright);
    s_aBright.setMaximum(255);
    s_aBright.setValue((int)aBright);
    s_bBright.setMinimum(0);
    s_bBright.setMaximum((int)aBright);
    s_bBright.setValue((int)bBright);
    
    s_aGray.setMinimum(0+(int)bGray);
    s_aGray.setMaximum(255-(int)cGray);
    s_aGray.setValue((int)aGray);
    s_bGray.setMinimum(0);
    s_bGray.setMaximum((int)aGray);
    s_bGray.setValue((int)bGray);
    s_cGray.setMinimum(0);
    s_cGray.setMaximum(255-(int)aGray);
    s_cGray.setValue((int)cGray);
  }
  
  private void resetMbf(){
    
    aDark = 70f;
    bDark = 57f;
    aBright = 180f;
    bBright = 57f;
    aGray = 127f;
    bGray = 57f;
    cGray = 57f;
    
    s_aDark.setMinimum(0);
    s_aDark.setMaximum(255-(int)bDark);
    s_aDark.setValue(70);
    s_bDark.setMinimum(0);
    s_bDark.setMaximum(255-(int)aDark);
    s_bDark.setValue(57);
    
    s_aBright.setMinimum(0+(int)bBright);
    s_aBright.setMaximum(255);
    s_aBright.setValue(180);
    s_bBright.setMinimum(0);
    s_bBright.setMaximum((int)aBright);
    s_bBright.setValue(57);
    
    s_aGray.setMinimum(0+(int)bGray);
    s_aGray.setMaximum(255-(int)cGray);
    s_aGray.setValue(127);
    s_bGray.setMinimum(0);
    s_bGray.setMaximum((int)aGray);
    s_bGray.setValue(57);
    s_cGray.setMinimum(0);
    s_cGray.setMaximum(255-(int)aGray);
    s_cGray.setValue(57);
  }
  
  //----------------metodi per la finestra dei singleton---------------
  
  private void showSingleton(){
    
    //creo i bottoni
    applySingleton = new JButton("Apply");
    applySingleton.addActionListener(this);
    resetSingleton = new JButton("Reset");
    resetSingleton.addActionListener(this);
    
    //creo i pannelli
    JPanel panelDARKER = new JPanel(new GridLayout(2,1));
    JPanel panelBRIGHTER = new JPanel(new GridLayout(2,1));
    JPanel panelGRAY = new JPanel(new GridLayout(2,1));
    JPanel panelApply_Reset = new JPanel();
    graphicSingleton = new PlotPanel(this.plotSingleton().getImage());
    
    //aggiungo un bordo ai pannelli
    panelDARKER.setBorder(BorderFactory.createTitledBorder("Singleton function DARKER"));
    panelGRAY.setBorder(BorderFactory.createTitledBorder("Singleton function GRAY"));
    panelBRIGHTER.setBorder(BorderFactory.createTitledBorder("Singleton function BRIGHTER"));
    
    //creo le etichette
    l_DARKER = new JLabel();
    l_GRAY = new JLabel();
    l_BRIGHTER = new JLabel();
    
    this.setLabelSingleton();
    
    //aggiungo le etichette dei campi
    panelDARKER.add(l_DARKER);
    panelGRAY.add(l_GRAY);
    panelBRIGHTER.add(l_BRIGHTER);
    
    //creo gli slider
    s_DARKER = new JSlider(JSlider.HORIZONTAL);
    s_GRAY = new JSlider(JSlider.HORIZONTAL);
    s_BRIGHTER = new JSlider(JSlider.HORIZONTAL);
    
    this.setSliderSingleton();
    
    s_DARKER.addChangeListener(this);
    s_GRAY.addChangeListener(this);
    s_BRIGHTER.addChangeListener(this);
    
    //aggiungo gli slider
    panelDARKER.add(s_DARKER);
    panelGRAY.add(s_GRAY);
    panelBRIGHTER.add(s_BRIGHTER);
    panelApply_Reset.add(applySingleton);
    panelApply_Reset.add(resetSingleton);
    
    //creo la finestra per le mbf
    windowSingleton = new JFrame("Singleton function - Fuzzy Contrast Enhancement");
    windowSingleton.setSize(700,450);
    windowSingleton.setLocation(300,200);
    windowSingleton.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    windowSingleton.setLayout(new FlowLayout());
    Container contentPane=windowSingleton.getContentPane();
    
    contentPane.add(graphicSingleton);
    contentPane.add(panelDARKER);
    contentPane.add(panelGRAY);
    contentPane.add(panelBRIGHTER);
    contentPane.add(panelApply_Reset);
    
    windowSingleton.setVisible(true);
  }
  
  private void setLabelSingleton(){
    
    l_DARKER.setText("DARKER = "+(int)DARKER);
    l_GRAY.setText("GRAY = "+(int)GRAY);
    l_BRIGHTER.setText("BRIGHTER = "+(int)BRIGHTER);
  }
  
  private void setSliderSingleton(){
    
    s_DARKER.setMinimum(0);
    s_DARKER.setMaximum(255);
    s_DARKER.setValue((int)DARKER);
    
    s_GRAY.setMinimum(0);
    s_GRAY.setMaximum(255);
    s_GRAY.setValue((int)GRAY);
    
    s_BRIGHTER.setMinimum(0);
    s_BRIGHTER.setMaximum(255);
    s_BRIGHTER.setValue((int)BRIGHTER);
  }
  
  private void resetSingleton(){
    
    DARKER = 0;
    GRAY = 127;
    BRIGHTER = 255;
    
    s_DARKER.setMinimum(0);
    s_DARKER.setMaximum(255);
    s_DARKER.setValue(0);
    
    s_GRAY.setMinimum(0);
    s_GRAY.setMaximum(255);
    s_GRAY.setValue(127);
    
    s_BRIGHTER.setMinimum(0);
    s_BRIGHTER.setMaximum(255);
    s_BRIGHTER.setValue(255);
  }
  
  //metodi per il controllo degli eventi
  public void stateChanged(ChangeEvent e){
    
    
    Object source=e.getSource();
    
    if(source==s_aDark){
      
      aDark = s_aDark.getValue();
      
      setSliderMbf();
      setLabelMbf();
      graphicMbf.updateImage(plotMbf().getImage());
    }
    
    if(source==s_bDark){
      
      bDark = s_bDark.getValue();
      
      setSliderMbf();
      setLabelMbf();
      graphicMbf.updateImage(plotMbf().getImage());
    }
    
    if(source==s_aGray){
      
      aGray = s_aGray.getValue();
      
      setSliderMbf();
      setLabelMbf();
      graphicMbf.updateImage(plotMbf().getImage());
    }
    
    if(source==s_bGray){
      
      bGray = s_bGray.getValue();
      
      setSliderMbf();
      setLabelMbf();
      graphicMbf.updateImage(plotMbf().getImage());
    }
    
    if(source==s_cGray){
      
      cGray = s_cGray.getValue();
      
      setSliderMbf();
      setLabelMbf();
      graphicMbf.updateImage(plotMbf().getImage());
    }
    
    if(source==s_aBright){
      
      aBright = s_aBright.getValue();
      
      setSliderMbf();
      setLabelMbf();
      graphicMbf.updateImage(plotMbf().getImage());
    }
    
    if(source==s_bBright){
      
      bBright = s_bBright.getValue();
      
      setSliderMbf();
      setLabelMbf();
      graphicMbf.updateImage(plotMbf().getImage());
    }
    
    if(source==s_DARKER){
      
      DARKER = s_DARKER.getValue();
      
      setSliderSingleton();
      setLabelSingleton();
      graphicSingleton.updateImage(plotSingleton().getImage());
    }
    
    if(source==s_GRAY){
      
      GRAY = s_GRAY.getValue();
      
      setSliderSingleton();
      setLabelSingleton();
      graphicSingleton.updateImage(plotSingleton().getImage());
    }
    
    if(source==s_BRIGHTER){
      
      BRIGHTER = s_BRIGHTER.getValue();
      
      setSliderSingleton();
      setLabelSingleton();
      graphicSingleton.updateImage(plotSingleton().getImage());
    }
    
  }
  
  public void actionPerformed(ActionEvent e){
    
    Object source=e.getSource();
    
    if(source==applyMbf){
      ip_fce.copyBits(ip_orig,0,0,Blitter.COPY);
      im = (byte[]) ip_fce.getPixels();
      this.process();
      im_fce.updateAndDraw();
    }
    
    if(source==resetMbf){
      this.resetMbf();
      setSliderMbf();
      setLabelMbf();
      graphicMbf.updateImage(plotMbf().getImage());
    }
    
    if(source==applySingleton){
      ip_fce.copyBits(ip_orig,0,0,Blitter.COPY);
      im = (byte[]) ip_fce.getPixels();
      this.process();
      im_fce.updateAndDraw();
    }
    
    if(source==resetSingleton){
      this.resetSingleton();
      setSliderSingleton();
      setLabelSingleton();
      graphicSingleton.updateImage(plotSingleton().getImage());
    }
    
  }
  
  public void windowClosing(WindowEvent e){
    
    windowMbf.setVisible(false);
    windowSingleton.setVisible(false);
    windowRules.setVisible(false);
  }
}